/*
 * Copyright (c) 2002 by Columbia University. All rights reserved.
 */

#include <stdio.h>     /* setvbuf() */
#include <stdarg.h>    /* vsnprintf */
#include <stdlib.h>    /* abort() */
#include <string.h>    /* strerror() */
#include <sys/time.h>  /* gettimeofday() */
#include <errno.h>     /* errno */
#include <syslog.h>    /* system log, unix */
#include <time.h>      /* strftime() */


FILE *debugfile = NULL; /* Name of debug/trace output file */

/*
 * debug()
 * Print debug message to stdout or log file.
 * Addopted and modified from cinema library
*/
void debug(const char *function_name, const char *format, ...)
{
  static int init_flag = 0;
  struct timeval tv;
  va_list ap;

  /* Default debug trace goes to stdout */
  if (debugfile == NULL) {
    debugfile = stdout;
  }

  va_start(ap, format);
  
  if (!init_flag) {
    setvbuf(debugfile, NULL, _IOLBF, 0);
    init_flag = 1;
  }

  gettimeofday(&tv, NULL);

  /* print out name of function causing error */
  if (function_name != NULL) {
    (void) fprintf(debugfile, "%9.3f %s: ", 
      tv.tv_sec + tv.tv_usec/1e6, function_name);
  }
  
  /*
  if (function_name != NULL) {
    (void) fprintf(debugfile, "%s: ", function_name);
  }
  */


  /* print out remainder of message */
  (void) vfprintf(debugfile, format, ap);

  if (debugfile != stdout) {
    fflush(debugfile);
  }
  
  va_end(ap);
} /* debug */
