  /***
   * Copyright (c) 2002 by Columbia University. All rights reserved.
   *
   * To run this program, you have to install OpenSLP package that 
   * can be downloaded from www.openslp.org.
   * For demonstration, use the slptool function in OpenSLP package
   * to register the following server.  This function may located 
   * in openslp-x.x.x/slptool/ directory.  
   * The following is the example from registration.
   *
   * ./slptool register service:printer://marta.cs.columbia.edu "(speed=200)"
   * ./slptool register service:printer://paris.clic.cs.columbia.edu "(speed=10)"
   * ./slptool register service:iptel-gw://baghdad.clic.cs.columbia.edu "(iptel-gw-prefix-list=1212,4930,8610),(iptel-gw-total-capacity=300),(iptel-gw-remaining-capacity=10),(iptel-gw-cost-list=1212:5,1:10,:20)"
   * ./slptool register service:iptel-gw://dehli.clic.cs.columbia.edu "(iptel-gw-prefix-list=4930,1212),(iptel-gw-total-capacity=30),(iptel-gw-remaining-capacity=10),(iptel-gw-cost-list=1212:5,1:10,:20)"
   * ./slptool register service:iptel-gw://cairo.clic.cs.columbia.edu "(iptel-gw-total-capacity=300),(iptel-gw-remaining-capacity=100)"
   * ./slptool register service:iptel-gw://munchen.clic.cs.columbia.edu "(iptel-gw-prefix-list=1212,4930,8610),(iptel-gw-total-capacity=300),(iptel-gw-remaining-capacity=10),(iptel-gw-cost-list=1212:5,1:10,:10)"
   * ./slptool register service:iptel-gw://jakarta.clic.cs.columbia.edu "(iptel-gw-prefix-list=1212,4930,8610),(iptel-gw-total-capacity=30),(iptel-gw-remaining-capacity=10),(iptel-gw-cost-list=1212:10,1:10,:10)"
   *
   ***/

#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <unistd.h>


#include "gwloc.h"

#define MAXLINE 10000

static void *doit(void *);
static void gwquery(int sockfd);

int main(int argc, char **argv)
{
  int         listenfd, connfd;
  pthread_t   tid;
  socklen_t   addrlen, len;
  struct sockaddr *cliaddr;
  struct sockaddr_in servaddr;

  if (argc !=2) {
    printf("usage: demo2cli <port>\n");
    exit(0);
  }


  listenfd = socket(AF_INET, SOCK_STREAM, 0);

  bzero(&servaddr, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servaddr.sin_port = htons(atoi(argv[1]));

  bind(listenfd, (struct sockaddr *) &servaddr, sizeof(servaddr));

  listen(listenfd, 1024);

  cliaddr = (struct sockaddr *) malloc(addrlen);


  for(;;) {
    len = addrlen;
    connfd = accept(listenfd, cliaddr, &len);
    pthread_create(&tid, NULL, &doit, (void *) connfd);
    printf("demo2srv standby\n");
  }
}

static void *
doit(void *arg)
{
  pthread_detach(pthread_self());
  gwquery((int) arg);
  close ((int) arg);
  return (NULL);
}

static void gwquery(int sockfd) {
  char msg[MAXLINE], send_msg[MAXLINE];
  char *p_strtok_r;
  int n;
  char *query, *filter;
  int i, j;
  gwloc_t *GwLoc;
  char **result;
  int *lifetime;
  int size, err;
  char bestmatch[] = "iptel-gw-cost-list=h,c=l";

  n = recv(sockfd, msg, MAXLINE, 0);
  if (n >= 0) {
    msg[n] = '\0';
    printf("msg= %s\n", msg);
    query = strtok_r(msg, "|", &p_strtok_r);
    filter = strtok_r(NULL, "|", &p_strtok_r);

    printf("query= %s\n", query);
    printf("filter= %s\n", filter);
        
    GwLoc = (gwloc_t *) malloc(sizeof(gwloc_t));
    err = Query(GwLoc, msg, filter);
    size = GetSize(GwLoc);
    result = (char **) malloc(sizeof(char *) * size);
    lifetime = (int *) malloc(sizeof(int) * size);
    GetResult(GwLoc, result, lifetime);

    j = 0;
    for (i = 0; i < size; i++) {
      j += sprintf(send_msg + j, "%s|", result[i]);
    }
    

    printf("Sending: %s\n", send_msg);
    n = send(sockfd, send_msg, strlen(send_msg), 0);
    printf("n= %d\n", n);

  }
}

