/*
 * Copyright (c) 2002 by Columbia University. All rights reserved.
 */

#include <iostream.h>
#include <stdlib.h>
#include "gwloc.h"

/*
 * Query()
 * API for user to do query to OpenSLP
 */
extern int Query(gwloc_t *GwLoc, char *msg_input, char *best_match_filter)
{
  int err;

  err = GW_SLP_Tool(GwLoc, msg_input, best_match_filter);

  return err;
} /* Query */

/*
 * GetSize()
 * Get the size of the array after doing Query()
 */
extern int GetSize(gwloc_t *GwLoc)
{
  if (GwLoc->parser.queryid == NULL) {
    printf("Do input Query first\n");
    return -1;
  }
  if (strcmp(GwLoc->parser.queryid, "SrvRqst") == 0)
  {
    return GwLoc->Srvsize;
  }
  if (strcmp(GwLoc->parser.queryid, "SrvTypeRqst") == 0)
  {
    return GwLoc->SrvTypesize;
  }
  if (strcmp(GwLoc->parser.queryid, "AttrRqst") == 0)
  {
    return GwLoc->Attrsize;
  }
  if (strcmp(GwLoc->parser.queryid, "PhoneNumber") == 0)
  {
    return GwLoc->Srvsize;
  }

  return 0;

} /* GetSize */

/*
 * GetResult()
 * Get the result of the query after doing Query()
 */
extern void GetResult(gwloc_t *GwLoc, char **result, int *lifetime)
{
  int i;

  if (GwLoc->parser.queryid == NULL) {
    result = NULL;
    lifetime = NULL;
    return;
  }

  if (strcmp(GwLoc->parser.queryid, "SrvRqst") == 0)
  {
    for (i = 0; i < GwLoc->Srvsize; i++)
    {
      result[i] = strdup(GwLoc->Srv[i]);
    }

    for (i = 0; i < GwLoc->Srvsize; i++)
    {
      lifetime[i] = GwLoc->Lifetime[i];
    }
    return;
  }
  if (strcmp(GwLoc->parser.queryid, "AttrRqst") == 0)
  {
    for (i = 0; i < GwLoc->Attrsize; i++)
    {
      result[i] = strdup(GwLoc->Attr[i]);
    }

    lifetime = NULL;


    return;
  }
  if (strcmp(GwLoc->parser.queryid, "SrvTypeRqst") == 0)
  {
    for (i = 0; i < GwLoc->SrvTypesize; i++)
    {
      result[i] = strdup(GwLoc->SrvType[i]);
    }

    lifetime = NULL;

    return;
  }
  if (strcmp(GwLoc->parser.queryid, "PhoneNumber") == 0)
  {
    for (i = 0; i < GwLoc->Srvsize; i++)
    {
      result[i] = strdup(GwLoc->Srv[i]);
    }

    for (i = 0; i < GwLoc->Srvsize; i++)
    {
      lifetime[i] = GwLoc->Lifetime[i];
    }

    return;
  }

} /* GetResult */

/* 
 * FreeGwLoc()
 * Free the allocated gwloc_t data structure
 */
extern void FreeGwLoc(gwloc_t *GwLoc)
{
  ParserFree(&GwLoc->parser);
  free(GwLoc->Output);
  free(GwLoc->Srv);
  free(GwLoc->SrvType);
  free(GwLoc->Attr);
  free(GwLoc);
} /* FreeGwLoc */
