/*
 * Copyright (c) 2002 by Columbia University. All rights reserved.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <slp.h>
#include <string.h>
#include <ctype.h>  /* toascii*/ 
#include <math.h>   /* fasb() */ 
#include <pthread.h>/* mutex  */


#ifndef	LIBGWLOC_H
#define	LIBGWLOC_H

#define MAXOUTPUT   100

#define GWLOC_BAD_QUERY   -27

/*
 * parser_t Structure 
 * used to store the user query as well as the result of parsed-query
 */
typedef struct parser {
  /* SLP URL as in draft-zhao-slp-url-00.txt */
  char   *slp_url;       /* "slp://" [domain-name] "/" query-desc */
  char   *domain_name;   /* hostname from section 5 of RFC 1738 */
  char   *query_desc;
  char   *queryid;       /* "SrvRqst" / "SrvTypeRqst" / "AttrRqst" */
  char   *scope;         /* scope-list from section 6.4.1 of RFC 2608 */
  char   *filter;        /* srvfilter / srvtypefilter / attrfilter */
  char   *srvfilter;     /* "type=" servicetype [";pred=" predicate]
                                         [";spi=" spistr] */
  char   *servicetype;   /* service type defined in section 4 of RFC 2608 */
  char   *predicate;     /* predicate from section 8.1 of RFC 2608 */
  char   *spistr;        /* SPI defined in section 9.2 of RFC 2608 */
  char   *srvtypefilter; /* "na=" nalenstr */
  char   *nalenstr;      /* "0" / "-1" / nastr */
  char   *nastr;         /* naming-auth from section 2.1 of RFC 2609 [5] */
  char   *attrfilter;    /* "url=" url [";tag=" taglist] [";spi=" spistr] */
  char   *url;           /* URL defined in section 4 of RFC 2608 */
  char   *taglist;       /* tag-list from section 10.3 of RFC 2608 */
  char   *extlist;       /* extension / extension ";" extlist */
  char   *extension;     /* attrext / selectext / sortext */
  char   *attrext;       /* "ext=attr" */
  char   *selectext;     /* "ext=select;" upper-bound */
  char   *upper_bound;   /* intval from section 5 of RFC 2608 */
  char   *sortext;       /* "ext=sort;" sort-order ";" sort-key-list */

  /* TELP NUMBER */
  char   *telp_number;   /* Telp. Number */
  char   *country_code;
  char   *area_code;

} parser_t;

/*
 * twloc_t structure 
 * stores all information as well as the result of the query for a specific user
 */
typedef struct gwloc {
  parser_t        parser;   /* To hold the parsed-query */
  char          **Output;   /* Generic result from OpenSLP Library */
  int           numOutput;  /* Generic number of result from OpenSLP Library */
  int           *Lifetime;  /* SA lifetimes */
  char          **Srv;      /* Server URL used by SrvRqst and Phone number query */
  char          **Attr;     /* Attributes, used by AttrRqst */
  char          **SrvType;  /* Service provided by DA, used by SrvTypeRqst */
  int           Srvsize;    /* Size of Srv */
  int           Attrsize;   /* Size of Attr */
  int           SrvTypesize;/* Size of SrvType */
  int           mult_factor;  /* multiplier factor, to keep track */ 
                              /* the dynamic memory alloc */
} gwloc_t;

/*
 * Structure processing Best Macth Gateway Query
 */
typedef struct t_attrlist {
  char    *srvname;
  int     lifetime;
  char    **srvattr;
  int     srvattrsize;
} ATTRLIST;


extern void ParserInit(parser_t *gw_slp);
extern void ParserFree(parser_t *gw_slp);
extern void ParseSlpUrl(parser_t *gw_slp, char *slp_url);
extern void ParsePhoneNumber(parser_t *gw_slp, char *phone_number);

void debug(const char *function_name, const char *format, ...);
extern int GW_SLP_Tool(gwloc_t *GwLoc, char *Query, char *BestMatchFilter);



#endif
