/*
 * Copyright (c) 2002 by Columbia University. All rights reserved.
 */

#include "libgwloc.h"

/*
 * ParserInit()
 * Parsing Initialization
 */ 
void ParserInit(parser_t *parser)
{

  parser->slp_url = NULL;
  parser->domain_name = NULL;
  parser->query_desc = NULL;
  parser->queryid = NULL;
  parser->scope = NULL;
  parser->filter = NULL;
  parser->srvfilter = NULL;
  parser->servicetype = NULL;
  parser->predicate = NULL;
  parser->spistr = NULL;
  parser->srvtypefilter = NULL;
  parser->nalenstr = NULL;
  parser->nastr = NULL;
  parser->attrfilter = NULL;
  parser->url = NULL;
  parser->taglist = NULL;
  parser->extlist = NULL;
  parser->extension = NULL;
  parser->attrext = NULL;
  parser->selectext = NULL;
  parser->upper_bound = NULL;
  parser->sortext = NULL;

  parser->telp_number = NULL;
  parser->area_code = NULL;
  parser->country_code = NULL;

} /* ParserInit */

/* 
 * ParserFree()
 * Free parser_t data type
 */
void ParserFree(parser_t *parser)
{
  free(parser->slp_url);
  free(parser->domain_name);
  free(parser->query_desc);
  free(parser->queryid);
  free(parser->scope);
  free(parser->filter);
  free(parser->srvfilter);
  free(parser->servicetype);
  free(parser->predicate);
  free(parser->spistr);
  free(parser->srvtypefilter);
  free(parser->nalenstr);
  free(parser->nastr);
  free(parser->attrfilter);
  free(parser->url);
  free(parser->taglist);
  free(parser->extlist);
  free(parser->extension);
  free(parser->attrext);
  free(parser->selectext);
  free(parser->upper_bound);
  free(parser->sortext);

  free(parser->telp_number);
  free(parser->country_code);
  free(parser->area_code);

} /* ParserFree */

/*
 * GW_SLP_ParseSrvfilter()
 * Parsing Srvfilter
 */
static void GW_SLP_ParseSrvfilter(parser_t *parser, char *srvfilter)
{
  char *servicetype, *predicate, *spistr;
  char *pdest;
  int loc;
  char *m;
  char *p_strtok_r;

  m = strdup(srvfilter);

  servicetype = strtok_r(m + 5, ";", &p_strtok_r);
  parser->servicetype = strdup(servicetype);

  if (strtok_r(NULL, ";", &p_strtok_r) != NULL) {
    
    m = strdup(srvfilter);
    pdest = strstr(m, "pred=");
    if (pdest != NULL) {
      loc = pdest - m + 1;
      predicate = strtok_r(m + loc + 4 , ";", &p_strtok_r);
      parser->predicate = strdup(predicate);
    }

    m = strdup(srvfilter);
    pdest = strstr(m, "spi=");
    if (pdest != NULL) {
      loc = pdest - m + 1;
      spistr = strtok_r(m + loc + 3, ";", &p_strtok_r);
      parser->spistr = strdup(spistr);
    }

  }

  free(m);

} /* GW_SLP_ParseSrvfilter */

/*
 * GW_SLP_ParseSrvtypefilter()
 * Parsing SrvTypefilter
 */
static void GW_SLP_ParseSrvtypefilter(parser_t *parser, char *srvtypefilter)
{

  parser->srvtypefilter = strdup(srvtypefilter);

  parser->nalenstr = strdup(srvtypefilter + 3);

  if ((strcmp(parser->nalenstr, "0") != 0) && (strcmp(parser->nalenstr, "-1") != 0)) {
    parser->nastr = strdup(parser->nalenstr);
  }

} /* GW_SLP_ParseSrvtypefilter */

/*
 * GW_SLP_ParseAttrfilter()
 * Parsing Attrfilter
 */
static void GW_SLP_ParseAttrfilter(parser_t *parser, char *attrfilter)
{
  char *url, *taglist, *spistr;
  char *pdest;
  int loc;
  char *m;
  char *p_strtok_r;

  m = strdup(attrfilter);

  url = strtok_r(m + 4, ";", &p_strtok_r);
  parser->url = strdup(url);

  if (strtok_r(NULL, ";", &p_strtok_r) != NULL) {
    
    m = strdup(attrfilter);
    pdest = strstr(m, "tag=");
    if (pdest != NULL) {
      loc = pdest - m + 1;
      taglist = strtok_r(m + loc + 3 , ";", &p_strtok_r);
      parser->taglist = strdup(taglist);
    }

    m = strdup(attrfilter);
    pdest = strstr(m, "spi=");
    if (pdest != NULL) {
      loc = pdest - m + 1;
      spistr = strtok_r(m + loc + 3, ";", &p_strtok_r);
      parser->spistr = strdup(spistr);
    }

  }

  free(m);

} /* GW_SLP_ParseAttrfilter */

/*
 * GW_SLP_ParseFilter()
 * Parsing the Filter
 */
static void GW_SLP_ParseFilter(parser_t *parser, char *filter)
{
  char *m;

  m = strdup(filter);

  if (strncmp(filter, "type=", 5) == 0) {
    parser->srvfilter = strdup(m);
    GW_SLP_ParseSrvfilter(parser, parser->srvfilter);
    return;
  }
  if (strncmp(filter, "na=", 3) == 0) {
    parser->srvtypefilter = strdup(m);
    GW_SLP_ParseSrvtypefilter(parser, parser->srvtypefilter);
    return;
  }
  if (strncmp(filter, "url=", 4) == 0) {
    parser->attrfilter = strdup(m);
    GW_SLP_ParseAttrfilter(parser, parser->attrfilter);
    return;
  }

  free(m);
} /* GW_SLP_ParseFilter */

/*
 * GW_SLP_ParseExtlist()
 * Parsing Extension List
 */
static void GW_SLP_ParseExtlist(parser_t *parser, char *extlist)
{
  char *m;

  m = strdup(extlist);

  free(m);

} /* GW_SLP_ParseExtlist */

/*
 * GW_SLP_ParseQueryDesc()
 * Parsing Query Description
 */
static void GW_SLP_ParseQueryDesc(parser_t *parser, char *query_desc)
{
  char *queryid, *scope, *filter, *extlist;
  int len;
  char *m;
  char *p_strtok_r;

  m = strdup(query_desc);

  queryid = strtok_r(m, "?", &p_strtok_r);
  parser->queryid = strdup(queryid);

  len = strlen(queryid);
  m = strdup(query_desc + len + 1);

  if (strncmp(m, "?", 1) == 0) {

    parser->scope = NULL;
    m = strdup(m + 1);

    filter = strtok_r(m, "?", &p_strtok_r);
    parser->filter = strdup(filter);
    
    extlist = strtok_r(NULL, "?", &p_strtok_r);

    if (extlist != NULL) {
      parser->extlist = strdup(extlist);
      GW_SLP_ParseExtlist(parser, parser->extlist);

    }
    else
    {
      parser->extlist = NULL;
    }

  } else {

    scope = strtok_r(m, "?", &p_strtok_r);
    parser->scope = strdup(scope);

    len = len + strlen(scope);
    m = strdup(query_desc + len + 1);

    filter = strtok_r(m, "?", &p_strtok_r);
    parser->filter = strdup(filter);
    
    extlist = strtok_r(NULL, "?", &p_strtok_r);

    if (extlist != NULL) {
      parser->extlist = strdup(extlist);
      GW_SLP_ParseExtlist(parser, parser->extlist);

    } else {
      parser->extlist = NULL;
    }

  }

  GW_SLP_ParseFilter(parser, filter);

  free(m);
  //free(queryid);
  //free(scope);
  //free(filter);
  //free(extlist);

} /* GW_SLP_ParseQueryDesc */

/*
 * ParseSlpUrl()
 * Parsing SLP URL
 */
void ParseSlpUrl(parser_t *parser, char *slp_url)
{
  char *domain_name;
  char *query_desc;
  char *m;
  char *p_strtok_r;

  m = strdup(slp_url);

  /* Get domain-name  and query-desc*/

  if (strncmp(m, "/", 1) == 0)
  {
    domain_name = "local host name";
    query_desc = strtok_r(m + 1, " ", &p_strtok_r); // add by 1 to avoid "/"
  } else {
    domain_name = strtok_r(m, "/", &p_strtok_r);
    query_desc = strtok_r(NULL, " ", &p_strtok_r);
  }


  parser->domain_name = strdup(domain_name);
  parser->query_desc = strdup(query_desc);

  GW_SLP_ParseQueryDesc(parser, query_desc);

  free(m);
  //free(domain_name);
  //free(query_desc);

} /* ParseSlpUrl */

/*
 * ParsePhoneNumber()
 * Parsing the Phone Number
 */
void ParsePhoneNumber(parser_t *parser, char *phone_number)
{
  char *country_code;
  char *area_code;
  char *m;
  char *p_strtok_r;

  m = strdup(phone_number);

  country_code = strtok_r(m + 4, "-", &p_strtok_r);
  area_code = strtok_r(NULL, "-", &p_strtok_r); 

  parser->country_code = strdup(country_code);
  parser->area_code = strdup(area_code);

  free(m);

} /* ParsePhoneNumber */
 




   
    
