/*
 * Copyright (c) 2002 by Columbia University. All rights reserved.
 */
#include "libgwloc.h"

/* mutexes defined individually for each query type */
pthread_mutex_t Srv_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t SrvType_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Attr_mutex = PTHREAD_MUTEX_INITIALIZER;

/* static variable to hold temporarily the result of callback */
static char **gSrvCallback;
char **gtmpSrvCallback;
int MultFactor;
static int gSrvCount;
static int *gLifetime;
static char *gAttrCallback = NULL;
static int gAttrCount;
static char *gSrvTypeCallback = NULL;
static int gSrvTypeCount;


#include "libgwloc.h"

/*
 * GW_SLP_SrvRqstCallback()
 * SrvRqst callback
 */
SLPBoolean GW_SLP_SrvRqstCallback( SLPHandle hslp, 
                                const char* srv, 
                                short unsigned int lifetime, 
                                SLPError errcode, 
                                void* cookie ) 
{ 
  char* cpy;
  int i;

  if (errcode == SLP_OK && *srv)
  { 
    cpy = strdup(srv);
    if (cpy)
    {
      /* Deallocate memory if exceeding the current contain of allocated memory */
      if (gSrvCount == (MultFactor * MAXOUTPUT))
      {
        MultFactor++;
        gtmpSrvCallback = (char **) malloc(sizeof(char *) * (MultFactor * MAXOUTPUT));
        for (i = 0; i < gSrvCount; i++)
        {
          gtmpSrvCallback[i] = strdup(gSrvCallback[i]);
        }
        free(gSrvCallback);
        gSrvCallback = (char **) malloc(sizeof(char *) * (MultFactor * MAXOUTPUT));
        for (i = 0; i < gSrvCount; i++)
        {
          gSrvCallback[i] = strdup(gtmpSrvCallback[i]);
        }
        free(gtmpSrvCallback);
      }

      gSrvCallback[gSrvCount] = strdup(cpy);
      debug("GW_SLP_SrvRqstCallback(LOG)", "gSrvCallback= %s\n", gSrvCallback[gSrvCount]);
      gLifetime[gSrvCount] = lifetime;
      debug("GW_SLP_SrvRqstCallback(LOG)", "gLifetime= %d\n", gLifetime[gSrvCount]);
      gSrvCount++;
      free(cpy);
    }
  } 
  
  return SLP_TRUE;
} /* GW_SLP_SrvRqstCallback */

/*
 * GW_SLP_AttrRqstCallback()
 * AttrRqst callback
 */
SLPBoolean GW_SLP_AttrRqstCallback(SLPHandle hslp, 
                           const char* cbattrlist, 
                           SLPError errcode, 
                           void* cookie ) 
{

  debug("GW_SLP_AttrRqstCallback(LOG)", "cbattrlist= %s\n", cbattrlist);
  if (errcode == SLP_OK)
  {
    gAttrCallback = strdup(cbattrlist);
  }
  
  return SLP_TRUE;

} /* GW_SLP_AttrRqstCallback */

/*
 * GW_SLP_ConvertToLDAPv3()
 * Convert URL to LDAPvv3 standard
 */
int GW_SLP_ConvertToLDAPv3(char *input, int len, char **result)
{
  char *pinput;
  char op[2], *tmp1, *tmp2, *tmp3;
  int i;

  pinput = (char *) malloc(sizeof(char) * len);
  tmp1 = (char *) malloc(sizeof(char) * len);
  tmp2 = (char *) malloc(sizeof(char) * len);
  tmp3 = (char *) malloc(sizeof(char) * len);
  pinput = strdup(input);

  
  for (i = 0; i < len; i++) {
    if (pinput[i] == '%') {
      switch(pinput[i + 1])
      {
      case '3':
        {
          switch(pinput[i + 2])
          {
          case 'c':
            strcpy(op,"<");
            break;
          
          case 'e':
            strcpy(op,">");
            break;

          case 'd':
            strcpy(op,"=");
            break;
          }

          break;
        }
     
      case '7':
        {
          switch(pinput[i + 2])
          {
          case 'c':
            strcpy(op,"|");
            break;
          
          case 'e':
            strcpy(op,"~");
            break;

          }
          
          break; 
        }

        
      case '2':
        {
          switch(pinput[i + 2])
          {
          case '1':
            strcpy(op,"!");
            break;
          
          case '6':
            strcpy(op,"&");
            break;

          case 'a':
            strcpy(op,"*");
            break;

          }

          break; 
        }
      }

      strcpy(tmp1, pinput + i + 3);
      strncpy(tmp2, pinput, i);
      tmp2[i] = '\0';
      strcpy(tmp3 , tmp2);
      strcat(tmp3, op);
      strcat(tmp3, tmp1);
      pinput = strdup(tmp3);
      len = len - 2;

    }

  }
  
  *result = strdup(pinput);
  free(pinput); free(tmp1); free(tmp2); free(tmp3);
  return 0;

} /* GW_SLP_ConvertToLDAPv3 */

/* 
 * CopyATTRLIST()
 * Copy ATTRLIST structure used by GW_SLP_GetBestMatchGateway()
 */
void CopyATTRLIST(ATTRLIST *dest, ATTRLIST *src)
{
  int k;

  dest->srvname  = strdup(src->srvname);
  dest->lifetime = src->lifetime;
  dest->srvattrsize = src->srvattrsize;
  for (k = 0; k < src->srvattrsize; k++)
  {
    dest->srvattr[k] = strdup(src->srvattr[k]);
  }
} /* CopyATTRLIST */

/*
 * InitATTRLIST()
 * Initialize ATTRLIST used by GW_SLP_GetBestMatchGateway()
 */
void InitATTRLIST(ATTRLIST **init, int initsize, int attrsize)
{
  int i;

  *init = (ATTRLIST *) malloc(sizeof(ATTRLIST ) * (initsize));
  
  for (i = 0; i < (initsize); i++)
  {
    init[i]->srvattr = (char **) malloc(sizeof(char *) * attrsize);
  }
  
} /* InitATTRLIST */

/*
 * FreeATTRLIST()
 * Free ATTRLIST structure used by GW_SLP_GetBestMatchGateway()
 */
void FreeATTRLIST(ATTRLIST *attr)
{
  free(attr);

} /* FreeATTRLIST */

/*
 * GW_SLP_GetBestMatchGateway()
 * Perform best matching filtering according 'BestMatchFilter'.  
 * No error returned for bad filter, but only OpenSLP error returned
 */
int GW_SLP_GetBestMatchGateway(parser_t *parser, char **Output, int *Lifetime, 
                               int *numOutput, char *BestMatchFilter, char *prefix)
{
  SLPError  result;
  SLPHandle hslp;
  int i, ii, j, flag;
  char *p_strtok_r1, *p_strtok_r2, *p_strtok_r3, *p_strtok_r4;
  char *token, *n, *attrname, *attrvalue, *attribute, *srv, *token_prefix, *token_info;
  char *attrlistresult;
  float bestvalue, dist;
  ATTRLIST *db, *bestdb;
  int attrnum, dbnum, bestdbnum;
  char *bmf, *tok, *attrlist, *tmp, *pdb;
  
  attrnum = 0;
  attrlist = (char *)malloc(sizeof(char) * strlen(BestMatchFilter));
  bmf = strdup(BestMatchFilter);
  //tok = strtok_r(bmf, ",=", &p_strtok_r1);
  tok = strtok_r(bmf, "=", &p_strtok_r1);
  strcpy(attrlist, tok);
  strcat(attrlist, ",");
  //tok = strtok_r(NULL, ",=", &p_strtok_r1);
  tok = strtok_r(NULL, ",", &p_strtok_r1);
  while( tok != NULL )
  {
    attrnum++;
    //tok = strtok_r( NULL, ",=", &p_strtok_r1);   
    tok = strtok_r( NULL, "=", &p_strtok_r1);  
    if (tok != NULL) {
      strcat(attrlist, tok);
      strcat(attrlist, ",");
    }
    //tok = strtok_r( NULL, ",=", &p_strtok_r1);
    tok = strtok_r( NULL, ",", &p_strtok_r1);
  }
  debug("GW_SLP_GetBestMatchGateway(LOG)", "attrlist= %s, attrnum= %d\n", attrlist, attrnum);

  // Initialize the database to store servers and their attributes
  //InitATTRLIST(&db, *numOutput, attrnum);
  debug("GW_SLP_GetBestMatchGateway(LOG)", "Allocate for db, with size= %d\n", *numOutput);
  db = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (*numOutput));
  for (i = 0; i < *numOutput; i++)
  {
    db[i].srvattr = (char **) malloc(sizeof(char *) * attrnum);
  }
  
  
  // Get attributes for each server
  if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK)
  {
    for (i = 0; i < *numOutput; i++) 
    {
      srv = strdup((char *)Output[i]);
      debug("GW_SLP_GetBestMatchGateway(LOG)", "Get Attributes for Srv%d= %s\n", i, srv);
      pthread_mutex_lock(&Attr_mutex);

      debug("GW_SLP_GetBestMatchGateway(LOG)", "attrlist= %s\n", attrlist);

      /* Beware!!!  Petensial slowness is because calling this function... */
      debug("GW_SLP_GetBestMatchGateway(LOG)", "Call OpenSLP function: SLPFindAttrs\n");
      result = SLPFindAttrs(hslp, 
                  srv,
                  parser->scope,
                  attrlist,
                  GW_SLP_AttrRqstCallback,
                  NULL);
      debug("GW_SLP_GetBestMatchGateway(LOG)", "Finish calling, Callback returned\n");

      if (result != SLP_OK)
      {
        debug("GW_SLP_SrvRqst(ERROR)", "errorcode: %i\n", result);
        pthread_mutex_unlock(&Attr_mutex);
        return result;
      }

      /*
      if (gAttrCallback == NULL) {
        pthread_mutex_unlock(&Attr_mutex);
        continue;
      }
      */

      if (gAttrCallback == NULL)
      {
        pthread_mutex_unlock(&Attr_mutex);
        attrlistresult = NULL;

        db[i].srvname = strdup(srv);
        db[i].lifetime = Lifetime[i];
        db[i].srvattr = NULL;
        db[i].srvattrsize = 0;
        debug("GW_SLP_GetBestMatchGateway(LOG)", "attrlistresult= %s\n", attrlistresult);
        
      }

      if (gAttrCallback != NULL) 
      {
        attrlistresult = strdup((char *) gAttrCallback);
        pthread_mutex_unlock(&Attr_mutex);

        // Put into db
        db[i].srvname = strdup(srv);
        db[i].lifetime = Lifetime[i];
        j = 0;
        debug("GW_SLP_GetBestMatchGateway(LOG)", "attrlistresult= %s\n", attrlistresult);
        tmp = strtok_r( attrlistresult, ")" , &p_strtok_r2);
        db[i].srvattr[j] = strdup(tmp);
        j++;
        while( tmp != NULL )
        {
        
          tmp = strtok_r( NULL, ")" , &p_strtok_r2);
          if (tmp != NULL) {
            db[i].srvattr[j] = strdup(tmp + 1);
            j++;
          }
        }
        db[i].srvattrsize = j;
      }
    }
  }
  SLPClose(hslp);

  // Process db to get the best gateway
  j = 0;
  dbnum = *numOutput;
  bmf = strdup(BestMatchFilter);
  attribute = strtok_r(bmf, ",", &p_strtok_r1);
  debug("GW_SLP_GetBestMatchGateway(LOG)", "attribute= %s\n", attribute);
  while( attribute != NULL )
  {
    n = strdup(attribute);
    attrname = strtok_r(n, "=", &p_strtok_r2);
    debug("GW_SLP_GetBestMatchGateway(LOG)", "attrname= %s\n", attrname);
    attrvalue = strtok_r(NULL, "=", &p_strtok_r2);
    debug("GW_SLP_GetBestMatchGateway(LOG)", "attrvalue= %s\n", attrvalue);

    flag = 0;
    
    if ((strcmp(attrname, "iptel-gw-cost-list") == 0) || 
        (strcmp(attrname, "iptel-gw-asr-list ") == 0))
    {
      if (prefix != NULL) 
      {
        bestdbnum = dbnum;
        
        //InitATTRLIST(&bestdb, bestdbnum, attrnum);
        bestdb = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (bestdbnum));
        for (i = 0; i < bestdbnum; i++)
        {
          bestdb[i].srvattr = (char **) malloc(sizeof(char *) * attrnum);
        }
        
        for (i = 0; i < dbnum; i++)
        {

          if (db[i].srvattr[j] !=NULL)
          {

            pdb = strdup(db[i].srvattr[j]);
            token = strtok_r((char *) pdb, "=", &p_strtok_r3);
            token = strtok_r( NULL, ",", &p_strtok_r3);

            while( token != NULL )
            {
              debug("GW_SLP_GetBestMatchGateway(LOG)", "token= %s\n", token);
              debug("GW_SLP_GetBestMatchGateway(LOG)", "prefix= %s\n", prefix);

              token_prefix = strtok_r(token, ":)", &p_strtok_r4);
              token_info = strtok_r(NULL, ",", &p_strtok_r4);

              if ((strcmp(prefix, token_prefix) == 0) || (token_info == NULL))   
                // (token_info == NULL) for Processing for unspecified prefix
              {
                if (flag == 0) {
                  if (token_info == NULL) {
                    token_info = token_prefix;
                  }
                  bestvalue = atof(token_info);
                  dist = fabs(bestvalue - atof(attrvalue));    
                  bestdbnum = 0;
                  CopyATTRLIST(&bestdb[bestdbnum], &db[i]);
                  bestdbnum = bestdbnum+1;

                  flag = 1;   // So this initialization will not be used anymore during this cycle

                } else {

                  if (token_info == NULL) {
                    token_info = token_prefix;
                  }
                  if (atof(token_info) == bestvalue)
                  {
                    CopyATTRLIST(&bestdb[bestdbnum], &db[i]);
                    bestdbnum++;
                  } else {

                    //Find a gateway with the lowest value
                    if ((strcmp(attrvalue, "l") == 0) || (strcmp(attrvalue, "L") == 0))
                    {
                      if (bestvalue > atof(token_info)) {
                        bestvalue = atof(token_info);

                        FreeATTRLIST(bestdb);
                        bestdbnum = 0;

                        //InitATTRLIST(&bestdb, bestdbnum, attrnum);
                        bestdb = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (dbnum));
                        for (ii = 0; ii < bestdbnum; ii++)
                        {
                          bestdb[ii].srvattr = (char **) malloc(sizeof(char *) * attrnum);
                        }
                        

                        CopyATTRLIST(&bestdb[bestdbnum], &db[i]);

                        bestdbnum = bestdbnum + 1;
                      }
                    }

                    //Find a gateway with the highest value
                    if ((strcmp(attrvalue, "h") == 0) || (strcmp(attrvalue, "H") == 0)) 
                    {
                      if (bestvalue < atof(token_info)) {
                        bestvalue = atof(token_info);

                        FreeATTRLIST(bestdb);
                        bestdbnum = 0;

                        //InitATTRLIST(&bestdb, bestdbnum, attrnum);
                        
                        bestdb = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (dbnum));
                        for (ii = 0; ii < bestdbnum; ii++)
                        {
                          bestdb[ii].srvattr = (char **) malloc(sizeof(char *) * attrnum);
                        }
                        

                        CopyATTRLIST(&bestdb[bestdbnum], &db[i]);

                        bestdbnum = bestdbnum + 1;                      }
                    }

                    //Find a gateway with the most close value
                    if ((toascii(attrvalue[0]) >= 48) && (toascii(attrvalue[0]) <= 57))
                    {
                      if (fabs(atof(token_info) - atof(attrvalue)) < dist) {
                        bestvalue = atof(token_info);
                        dist = fabs(atof(token_info) - atof(attrvalue));

                        FreeATTRLIST(bestdb);
                        bestdbnum = 0;

                        //InitATTRLIST(&bestdb, bestdbnum, attrnum);
                        
                        bestdb = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (dbnum));
                        for (ii = 0; ii < bestdbnum; ii++)
                        {
                          bestdb[ii].srvattr = (char **) malloc(sizeof(char *) * attrnum);
                        }
                        

                        CopyATTRLIST(&bestdb[bestdbnum], &db[i]);

                        bestdbnum = bestdbnum + 1;                      
                      }
                    }
                  }
                }
              }
              
              token = strtok_r( NULL, ",", &p_strtok_r3);
            } //while( token != NULL )

          } // if (db[i]->srvattr[j] !=NULL)

        } // for (i = 0; i < dbnum; i++)

      } //if (prefix != NULL) 

      if (flag == 1)
      {
        dbnum = bestdbnum;
        FreeATTRLIST(db);

        //InitATTRLIST(&db, dbnum, attrnum);
        
        db = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (dbnum));
        for (ii = 0; ii < dbnum; ii++)
        {
          db[ii].srvattr = (char **) malloc(sizeof(char *) * attrnum);
        }
        
        
        for (i = 0; i < dbnum; i++) {
          CopyATTRLIST(&db[i], &bestdb[i]);
        }
        
        FreeATTRLIST(bestdb);

      } // if (flag == 1)

    }
    else
    {
      if (attrvalue != NULL)
      {
        // For H, h, L, or l
        if ((strcmp(attrvalue, "l") == 0) || (strcmp(attrvalue, "L") == 0) ||
          (strcmp(attrvalue, "h") == 0) || (strcmp(attrvalue, "H") == 0) ||
          (isdigit(attrvalue[0])))
        {

          bestdbnum = dbnum;
        
          //InitATTRLIST(&bestdb, bestdbnum, attrnum);
        
          bestdb = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (bestdbnum));
          for (i = 0; i < bestdbnum; i++)
          {
            bestdb[i].srvattr = (char **) malloc(sizeof(char *) * attrnum);
          }
        
        
          for (i = 0; i < dbnum; i++)
          {
            if (db[i].srvattr[j] !=NULL)
            {
              token = strtok_r((char *) db[i].srvattr[j], "=", &p_strtok_r3);
              token = strtok_r( NULL, ",", &p_strtok_r3);

              while( token != NULL )
              {

                if (flag == 0) {
                
                  bestvalue = atof(token);
                  dist = fabs(bestvalue - atof(attrvalue));    
                  bestdbnum = 0;

                  CopyATTRLIST(&bestdb[bestdbnum], &db[i]);
                  bestdbnum = bestdbnum+1;

                  flag = 1;   // So this initialization will not be used anymore during this cycle

                } else {

                  if (atof(token) == bestvalue)
                  {
                    CopyATTRLIST(&bestdb[bestdbnum], &db[i]);
                    bestdbnum++;
                  } else {

                    if ((strcmp(attrvalue, "l") == 0) || (strcmp(attrvalue, "L") == 0))
                    {
                      if (bestvalue > atof(token)) {
                        bestvalue = atof(token);

                        FreeATTRLIST(bestdb);
                        bestdbnum = 0;

                        //InitATTRLIST(&bestdb, bestdbnum, attrnum);
                      
                        bestdb = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (dbnum));
                        for (ii = 0; ii < bestdbnum; ii++)
                        {
                          bestdb[ii].srvattr = (char **) malloc(sizeof(char *) * attrnum);
                        }
                      

                        CopyATTRLIST(&bestdb[bestdbnum], &db[i]);

                        bestdbnum = bestdbnum + 1;
                      }
                    }

                    //Find a gateway with the highest value
                    if ((strcmp(attrvalue, "h") == 0) || (strcmp(attrvalue, "H") == 0)) 
                    {
                      if (bestvalue < atof(token)) {
                        bestvalue = atof(token);

                        FreeATTRLIST(bestdb);
                        bestdbnum = 0;

                        //InitATTRLIST(&bestdb, bestdbnum, attrnum);
                      
                        bestdb = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (dbnum));
                        for (ii = 0; ii < bestdbnum; ii++)
                        {
                          bestdb[ii].srvattr = (char **) malloc(sizeof(char *) * attrnum);
                        }
                      

                        CopyATTRLIST(&bestdb[bestdbnum], &db[i]);

                        bestdbnum = bestdbnum + 1;                      }
                    }

                    //Find a gateway with the most close value
                    if ((toascii(attrvalue[0]) >= 48) && (toascii(attrvalue[0]) <= 57))
                    {
                      if (fabs(atof(token) - atof(attrvalue)) < dist) {
                        bestvalue = atof(token);
                        dist = fabs(atof(token) - atof(attrvalue));

                        FreeATTRLIST(bestdb);
                        bestdbnum = 0;

                        //InitATTRLIST(&bestdb, bestdbnum, attrnum);
                      
                        bestdb = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (dbnum));
                        for (ii = 0; ii < bestdbnum; ii++)
                        {
                          bestdb[ii].srvattr = (char **) malloc(sizeof(char *) * attrnum);
                        }
                      

                        CopyATTRLIST(&bestdb[bestdbnum], &db[i]);

                        bestdbnum = bestdbnum + 1;                      
                      }
                    }
                  }
                }
              
                token = strtok_r( NULL, ",", &p_strtok_r3);
              } //while( token != NULL )

            } // if (db[i]->srvattr[j] !=NULL)

          } // for (i = 0; i < dbnum; i++)

          if (flag == 1)
          {
            dbnum = bestdbnum;
            FreeATTRLIST(db);

            //InitATTRLIST(&db, dbnum, attrnum);
          
            db = (ATTRLIST *) malloc(sizeof(ATTRLIST) * (dbnum));
            for (ii = 0; ii < dbnum; ii++)
            {
              db[ii].srvattr = (char **) malloc(sizeof(char *) * attrnum);
            }
          

            for (i = 0; i < dbnum; i++) {
              CopyATTRLIST(&db[i], &bestdb[i]);
            }
        
            FreeATTRLIST(bestdb);

          } // if (flag == 1)

        } // if ((strcmp(attrvalue, "l") == 0) .....
      
      } // // if (strcmp != NULL)


    } // else


    attribute = strtok_r(NULL, ",", &p_strtok_r1);
    j++;

  }

  *numOutput = dbnum;
  for (i = 0; i < *numOutput; i++)
  {
    Output[i] = strdup(db[i].srvname);
    Lifetime[i] = db[i].lifetime;
  }
  FreeATTRLIST(db);
  free(attrlist);


  
  return result;

} /* GW_SLP_GetBestMatchGateway */

/*
 * GW_SLP_GetPrefix()
 * 'SLPfilter' contains "iptel-gw-prefix-list"
 * This function tried to get the value of "iptel-gw-prefix-list"
 * the value is returned to 'searched_prefix'
 */
static void GW_SLP_GetPrefix(char *SLPfilter, char **searched_prefix)
{
  char *str1, *str2, *token, *value;
  int i;
  char *pstrtok_r;

  str1 = (char *) malloc(sizeof(char) * strlen(SLPfilter));
  strcpy(str1, SLPfilter);

  // Search the substring "iptel-gw-prefix-list"
  for (i = 0; i < (int) strlen(SLPfilter); i++)
  {
    if ((str1[i] == 'i') || (str1[i] == 'I')) {
      str2 = str1 + i;
      if (strncasecmp(str2, "iptel-gw-prefix-list", strlen("iptel-gw-prefix-list")) == 0) {
        // the substring is found, get the value
        token = strtok_r(str2, "=", &pstrtok_r);
        value = strtok_r(NULL, ")", &pstrtok_r);
        *searched_prefix = strdup(value);
        free(str1);
        return;
      }
    }
  }
  *searched_prefix = NULL;
  free(str1);

} /* GW_SLP_GetPrefix */

/*
 * GW_SLP_SrvRqst()
 * To do SrvRqst query
 */
int GW_SLP_SrvRqst(gwloc_t *GwLoc, char *BestMatchFilter) 

{
  SLPError  result;
  SLPHandle hslp;
  int i;
  void *pCookie;
  char *filter;
  char *searched_prefix_list;



  if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK)
  {

    pthread_mutex_lock(&Srv_mutex);

    debug("GW_SLP_SrvRqst(LOG)", "GwLoc->parser.predicate= %s\n", GwLoc->parser.predicate);
    
    if (GwLoc->parser.predicate != NULL)
    { 
      
      GW_SLP_ConvertToLDAPv3(GwLoc->parser.predicate, strlen(GwLoc->parser.predicate), &filter);
      debug("GW_SLP_SrvRqst(LOG)", "filter= %s\n", filter);

      gSrvCount = 0;
      MultFactor = GwLoc->mult_factor;

      /* Beware!!!  Petensial slowness is because calling this function... */
      debug("GW_SLP_SrvRqst(LOG)", "Call OpenSLP function: SLPFindSrvs()\n");
      result = SLPFindSrvs(hslp,
                           GwLoc->parser.servicetype,
                           GwLoc->parser.scope,
                           filter,                
                           GW_SLP_SrvRqstCallback,
                           pCookie);
      debug("GW_SLP_SrvRqst(LOG)", "Finish calling, Callback returned\n");

    } else {

      gSrvCount = 0;
      MultFactor = GwLoc->mult_factor;

      /* Beware!!!  Petensial slowness is because calling this function... */
      debug("GW_SLP_SrvRqst(LOG)", "Call OpenSLP function: SLPFindSrvs()\n");
      result = SLPFindSrvs(hslp,
                           GwLoc->parser.servicetype,
                           GwLoc->parser.scope,
                           0,                
                           GW_SLP_SrvRqstCallback,
                           pCookie);
      debug("GW_SLP_SrvRqst(LOG)", "Finish calling, Callback returned\n");
    }

    if (result != SLP_OK) 
    {
      debug("GW_SLP_SrvRqst(ERROR)", "errorcode: %i\n", result);
      pthread_mutex_unlock(&Srv_mutex);
      return result;
    }

    GwLoc->Output = (char **) malloc(sizeof(char *) * gSrvCount);
    GwLoc->Lifetime = (int *) malloc(sizeof(int) * gSrvCount);
    GwLoc->numOutput = gSrvCount;
    for (i = 0; i < GwLoc->numOutput; i++)
    {
      GwLoc->Output[i] = strdup(gSrvCallback[i]);
      GwLoc->Lifetime[i] = gLifetime[i];
      debug("GW_SLP_SrvRqst(LOG)", "output[%d]= %s\n", i, GwLoc->Output[i]);
      debug("GW_SLP_SrvRqst(LOG)", "lifetime[i]= %d\n", GwLoc->Lifetime[i]);
    }
    gSrvCount = 0;
    GwLoc->mult_factor = MultFactor;
    pthread_mutex_unlock(&Srv_mutex);

    SLPClose(hslp);

    debug("GW_SLP_SrvRqst(LOG)", "BestMatchFilter= %s, filter= %s\n", BestMatchFilter, filter);

    
    GW_SLP_GetPrefix(filter, &searched_prefix_list);

    if ((BestMatchFilter != NULL) && (searched_prefix_list != NULL)) {
      GW_SLP_GetBestMatchGateway(&GwLoc->parser, GwLoc->Output, GwLoc->Lifetime, 
        &GwLoc->numOutput, BestMatchFilter, searched_prefix_list);
    }
    
  }

  return result;
  
} /* GW_SLP_SrvRqst */

/*
 * GW_SLP_SrvTypeCallback()
 * Callback for SLPFindSrvTypes()
 */
SLPBoolean GW_SLP_SrvTypeCallback(SLPHandle hslp, 
                                const char* srv, 
                                SLPError errcode, 
                                void* cookie ) 
{ 
  char* cpy;

  if (errcode == SLP_OK && *srv)
  { 
    cpy = strdup(srv);
    if (cpy)
    {
      gSrvTypeCallback = strdup(cpy);
      gSrvTypeCount++;
      free(cpy);
    }
  } 
  
  return SLP_TRUE;
} /* GW_SLP_SrvTypeCallback */

/*
 * GW_SLP_SrvTypeRqst()
 * Query based on Service Type
 */
int GW_SLP_SrvTypeRqst(gwloc_t *GwLoc) 
{
  SLPError  result;
  SLPHandle hslp;
  int i;


  if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK)
  {
    
    pthread_mutex_lock(&SrvType_mutex);
    

    gSrvCount = 0;
    if (GwLoc->parser.nalenstr)
    {
      /* Beware!!!  Petensial slowness is because calling this function... */
      debug("GW_SLP_SrvTypeRqst(LOG)", "Call OpenSLP function: SLPFindSrvTypes()\n");
      result = SLPFindSrvTypes(hslp,
                             GwLoc->parser.nalenstr,
                             GwLoc->parser.scope,
                             GW_SLP_SrvTypeCallback,
                             0);
      debug("GW_SLP_SrvTypeRqst(LOG)", "Finish calling, Callback returned\n");
    }
    else
    
    {
      /* Beware!!!  Petensial slowness is because calling this function... */
      debug("GW_SLP_SrvTypeRqst(LOG)", "Call OpenSLP function: SLPFindSrvTypes()\n");
      result = SLPFindSrvTypes(hslp,
                               "*",   
                               GwLoc->parser.scope,
                               GW_SLP_SrvTypeCallback,
                               0);
      debug("GW_SLP_SrvTypeRqst(LOG)", "Finish calling, Callback returned\n");
    }

    if (result != SLP_OK)
    {
      debug("GW_SLP_SrvTypeRqst(ERROR)", "errorcode: %i\n", result);
      pthread_mutex_unlock(&SrvType_mutex);
      return result;
    }


    GwLoc->numOutput = gSrvTypeCount;
    GwLoc->Output = (char **) malloc(sizeof(char *) * gSrvTypeCount);

    for (i = 0; i < (GwLoc->numOutput); i++)
    {
      GwLoc->Output[i] = strdup((char *)gSrvTypeCallback);
    }

    gSrvTypeCount = 0;
    pthread_mutex_unlock(&SrvType_mutex);
    
    SLPClose(hslp);
    
  }
  return result;

} /* GW_SLP_SrvTypeRqst */

/*
 * GW_SLP_AttrRqst()
 * Query based on Attribute
 */
int GW_SLP_AttrRqst(gwloc_t *GwLoc)
{
  SLPError  result;
  SLPHandle hslp;
  char *attribute = NULL;
  char *thisurl;
  SLPSrvURL *ParsedUrl;

  thisurl = (char *) malloc(sizeof(char) * (strlen(GwLoc->parser.url) + 19));

  if (GwLoc->parser.attrext != NULL)
    attribute = GwLoc->parser.attrext + 4;  // 4 is to avoid "ext=" word

  debug("GW_SLP_AttrRqst(LOG)", "GwLoc->parser.url= %s\n", GwLoc->parser.url);
  SLPParseSrvURL(GwLoc->parser.url, &ParsedUrl);

  strcpy(thisurl, ParsedUrl->s_pcSrvType);
  strcat(thisurl, "://");
  strcat(thisurl, ParsedUrl->s_pcHost);

  
  if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK)
  {

    pthread_mutex_lock(&Attr_mutex);

    gAttrCount = 0;

    /* Beware!!!  Petensial slowness is because calling this function... */
    debug("GW_SLP_AttrRqst(LOG)", "Call OpenSLP function: SLPFindAttrs()\n");
    result = SLPFindAttrs(hslp, 
                      thisurl,
                      GwLoc->parser.scope,
                      attribute,
                      GW_SLP_AttrRqstCallback,
                      0);
    debug("GW_SLP_AttrRqst(LOG)", "Finish calling, Callback returned\n");

    if (result != SLP_OK)
    {
      debug("GW_SLP_AttrRqst(ERROR)", "errorcode: %i\n", result);
      pthread_mutex_unlock(&Attr_mutex);
      return result;
      
    }

    debug("GW_SLP_AttrRqst(LOG)", "gAttrCallback= %s\n", gAttrCallback);
    
    GwLoc->Output = (char **) malloc(sizeof(char *) * gAttrCount);
    GwLoc->Output[0] = strdup((char *)gAttrCallback);
    debug("GW_SLP_AttrRqst(LOG)", "GwLoc->Output[0]= %s\n", GwLoc->Output[0]);
    gAttrCount++;
    GwLoc->numOutput = gAttrCount;
    debug("GW_SLP_AttrRqst(LOG)", "GwLoc->numOutput= %d\n", GwLoc->numOutput);

    gAttrCallback = NULL;
    gAttrCount = 0;
    pthread_mutex_unlock(&Attr_mutex);

    SLPClose(hslp);

    
  }
  

  SLPFree(ParsedUrl);
  free(thisurl);

  return result;

} /* GW_SLP_AttrRqst */

/* GW_SLP_GatewayRqst()
 * Get gateway for phone number query
 */
int GW_SLP_GatewayRqst(gwloc_t *GwLoc, char *BestMatchFilter) 

{
  SLPError  result;
  SLPHandle hslp;
  int i;
  char *searched_prefix_list;
  char *filter;

  if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK)
  {

    /* Search the appropriate prefix-list */
    searched_prefix_list = (char *) malloc(sizeof(char) * (strlen(GwLoc->parser.country_code) + strlen(GwLoc->parser.area_code)));
    strcpy(searched_prefix_list, GwLoc->parser.country_code);
    strcat(searched_prefix_list, GwLoc->parser.area_code);

    // Construct LDAPv3 filter
    filter = (char *) malloc(sizeof(char) * (strlen(searched_prefix_list) + strlen("(searched_prefix_list=)")));
    strcpy(filter, "(iptel-gw-prefix-list="); 
    strcat(filter, searched_prefix_list);
    strcat(filter, ")");
    debug("GW_SLP_GatewayRqst(LOG)", "filter= %s\n", filter);

    pthread_mutex_lock(&Srv_mutex);

    gSrvCount = 0;
    MultFactor = GwLoc->mult_factor;

    /* Find all "iptel-gw" server */

    /* Beware!!!  Petensial slowness is because calling this function... */
    debug("GW_SLP_GatewayRqst(LOG)", "Call OpenSLP function: SLPFindSrvs()\n");
    result = SLPFindSrvs(hslp,
                         "iptel-gw",
                         0,
                         filter,                
                         GW_SLP_SrvRqstCallback,
                         0);
    debug("GW_SLP_GatewayRqst(LOG)", "Finish calling, Callback returned\n");

    if (result != SLP_OK) 
    {
      debug("GW_SLP_GatewayRqst(ERROR)", "errorcode: %i\n", result);
      pthread_mutex_unlock(&Srv_mutex);
      return result;
    }

    GwLoc->numOutput = gSrvCount;
    GwLoc->Output = (char **) malloc(sizeof(char *) * gSrvCount);
    GwLoc->Lifetime = (int *) malloc(sizeof(int) * gSrvCount);

    for (i = 0; i < GwLoc->numOutput; i++)
    {
      GwLoc->Output[i] = strdup(gSrvCallback[i]);
      GwLoc->Lifetime[i] = gLifetime[i];
      debug("GW_SLP_GatewayRqst(LOG)", "output[%d]= %s\n", i, GwLoc->Output[i]);
      debug("GW_SLP_GatewayRqst(LOG)", "lifetime[i]= %d\n", GwLoc->Lifetime[i]);
    }

    gSrvCount = 0;
    GwLoc->mult_factor = MultFactor;
    pthread_mutex_unlock(&Srv_mutex);

    SLPClose(hslp);
    

    debug("GW_SLP_GatewayRqst(LOG)", "BestMatchFilter= %s\n", BestMatchFilter);
    if (BestMatchFilter) {
      // Process Output
      GW_SLP_GetBestMatchGateway(&GwLoc->parser, GwLoc->Output, GwLoc->Lifetime, 
        &GwLoc->numOutput, BestMatchFilter, searched_prefix_list);
    }

    free(searched_prefix_list); free(filter);
    

  }

  return result;

} /* GW_SLP_GatewayRqst */

/*  
 * GetServerName()
 * Get the name of the server only from the SLPFindSrvs()
 */
char *GetServerName(char *url)
{
  char *servername;
  char *p_strtok_r;
  char *m;

  m = strdup((char *)url);

  servername = strtok_r(m, "//", &p_strtok_r);
  servername = strtok_r(NULL, "//", &p_strtok_r);

  return servername;
} /* GetServerName */

/* 
 * GW_SLP_GetSrv()
 * Get the server name for API purpose
 */
void GW_SLP_GetSrv(gwloc_t *GwLoc)
{
  int i;

  GwLoc->Srv = (char **) malloc(sizeof(char *) * (GwLoc->numOutput)); 
  for (i = 0; i < (GwLoc->numOutput); i++)
  {
    GwLoc->Srv[i] = strdup(GetServerName(GwLoc->Output[i]));
  }
  GwLoc->Srvsize = (GwLoc->numOutput);

} /* GW_SLP_GetSrv */

/* 
 * GW_SLP_GetSrvType()
 * Get Service Type for API purpose
 */
void GW_SLP_GetSrvType(gwloc_t *GwLoc)
{
  char *cpy, *pcpy;
  char *tmp1;
  char *tmp2;
  int n_type, i;
  char *p_strtok_r;

  // Find out the size of array
  n_type = 0;
  cpy = strdup(GwLoc->Output[0]);
  if (cpy) {
    tmp1 = tmp2 = cpy;
    tmp1 = strchr(tmp2, ',');
    while (tmp1) {
      *tmp1 = 0;
      n_type++;
      tmp1 ++;
      tmp2 = tmp1;
      tmp1 = strchr(tmp2, ',');
    }
    n_type++;
  }

  // Allocate memory
  GwLoc->SrvTypesize = n_type;
  GwLoc->SrvType = (char **) malloc(sizeof(char *) * (GwLoc->SrvTypesize));

  cpy = strdup(GwLoc->Output[0]);
  pcpy = strtok_r(cpy, ",", &p_strtok_r);
  GwLoc->SrvType[0] = strdup(pcpy + 8);  // service: is 8 words
  for (i = 1; i < (GwLoc->SrvTypesize); i++)
  {
    pcpy = strtok_r(NULL, ",", &p_strtok_r);
    GwLoc->SrvType[i] = strdup(pcpy + 8);
  }

} /* GW_SLP_GetSrvType */

/* 
 * GW_SLP_GetAttr()
 * Get the attribute list for API purpose
 */
void GW_SLP_GetAttr(gwloc_t *GwLoc)
{
    
  char *cpy, *pcpy;
  char *tmp1;
  char *tmp2;
  int n_attr, i;
  char *p_strtok_r;

  // Find out the size of array
  n_attr = 0;
  cpy = strdup(GwLoc->Output[0]);
  if (cpy) {
    tmp1 = tmp2 = cpy;
    tmp1 = strchr(tmp2, '(');
    while (tmp1) {
      *tmp1 = 0;
      n_attr++;
      tmp1 ++;
      tmp2 = tmp1;
      tmp1 = strchr(tmp2, '(');
    }
  }

  // Allocate memory
  GwLoc->Attrsize = n_attr;
  GwLoc->Attr = (char **) malloc(sizeof(char *) * (GwLoc->Attrsize)); 

  cpy = strdup(GwLoc->Output[0]);
  pcpy = strtok_r(cpy, ")", &p_strtok_r);
  free(GwLoc->Attr);
  GwLoc->Attr[0] = strdup(pcpy);  
  strcat(GwLoc->Attr[0], ")");
  debug("GW_SLP_GetAttr(LOG)", "Attr[0]= %s\n", GwLoc->Attr[0]);
  for (i = 1; i < (GwLoc->Attrsize); i++)
  {
    pcpy = strtok_r(NULL, ")", &p_strtok_r);
    GwLoc->Attr[i] =strdup(pcpy + 1);
    strcat(GwLoc->Attr[i], ")");
    debug("GW_SLP_GetAttr(LOG)", "Attr[%d]= %s\n", i, GwLoc->Attr[i]);
  }

} /* GW_SLP_GetAttr */

/*
 * GW_SLP_Tool()
 * Main function for doing query, interface to the API in gwloc.c
 */
extern int GW_SLP_Tool(gwloc_t *GwLoc, char *Query, char *BestMatchFilter)
{
  int i, err;

  // Don't know why, but have to do this, otherwise segmentation fault
  
  pthread_mutex_lock(&Attr_mutex);
  pthread_mutex_unlock(&Attr_mutex);

  pthread_mutex_lock(&SrvType_mutex);
  pthread_mutex_unlock(&SrvType_mutex);
  

  pthread_mutex_lock(&Srv_mutex);
  pthread_mutex_unlock(&Srv_mutex);

  gSrvCallback = (char **) malloc(sizeof(char *) * MAXOUTPUT);  
  gLifetime = (int *) malloc(sizeof(int) * MAXOUTPUT);
  GwLoc->mult_factor = 1;


  pthread_mutex_lock(&Srv_mutex);
  pthread_mutex_unlock(&Srv_mutex);


  ParserInit(&GwLoc->parser);


  if (strncmp(Query, "slp://", 6) == 0) 
  {

    GwLoc->parser.slp_url = strdup(Query);
    ParseSlpUrl(&GwLoc->parser, Query + 6);


    if (strcmp(GwLoc->parser.queryid, "SrvRqst") == 0) {
      err = GW_SLP_SrvRqst(GwLoc, BestMatchFilter);
      if (err == 0)
        GW_SLP_GetSrv(GwLoc);

      // Try to print
      /*
      for (i = 0; i < GwLoc->Srvsize; i++) {
        debug("GW_SLP_Tool(LOG)", "GwLoc->Srv= %s\n", GwLoc->Srv[i]);
      }
      */
      free(gSrvCallback);
      free(gLifetime);
      free(gAttrCallback);
      free(gSrvTypeCallback);

      return err;
    } 
  
    if (strcmp(GwLoc->parser.queryid, "SrvTypeRqst") == 0) {
      err = GW_SLP_SrvTypeRqst(GwLoc);
      if (err == 0)
        GW_SLP_GetSrvType(GwLoc);
      
      // Try to print
      /*
      for (i = 0; i < GwLoc->SrvTypesize; i++) {
        debug("GW_SLP_Tool(LOG)", "GwLoc->SrvType= %s\n", GwLoc->SrvType[i]);
      }
      */
      free(gSrvCallback);
      free(gLifetime);
      free(gAttrCallback);
      free(gSrvTypeCallback);

      return err;
    }

    if (strcmp(GwLoc->parser.queryid, "AttrRqst") == 0) {
      err = GW_SLP_AttrRqst(GwLoc);
      if (err == 0)
        GW_SLP_GetAttr(GwLoc);
      
      // Try to print
      
      for (i = 0; i < GwLoc->Attrsize; i++) {
        debug("GW_SLP_Tool(LOG)", "GwLoc->Attr= %s\n", GwLoc->Attr[i]);
      }
      
      
      free(gSrvCallback);
      free(gLifetime);
      free(gAttrCallback);
      free(gSrvTypeCallback);
      
      return err;
    }
  }

  if (strncmp(Query, "tel:", 4) == 0) 
  {
    debug("GW_SLP_Parser(LOG)", "Parsing Phone Number: %s\n", Query);
    ParsePhoneNumber(&GwLoc->parser, Query);

    GwLoc->parser.queryid = "PhoneNumber";
    GwLoc->parser.scope = NULL;

    err = GW_SLP_GatewayRqst(GwLoc, BestMatchFilter);
    if (err == 0)
      GW_SLP_GetSrv(GwLoc);
    
    // Try to print
    /*
    for (i = 0; i < GwLoc->Srvsize; i++) {
      debug("GW_SLP_Tool(LOG)", "GwLoc->Srv= %s\n", GwLoc->Srv[0]);
    }
    */
    
    free(gSrvCallback);
    free(gLifetime);
    free(gAttrCallback);
    free(gSrvTypeCallback);

    return err;

  }


  debug("GW_SLP_Tool(ERROR)", "Error, unknown query= %s\n", Query);

  ParserFree(&GwLoc->parser);
  free(gSrvCallback);
  free(gLifetime);
  free(gAttrCallback);
  free(gSrvTypeCallback);

  return GWLOC_BAD_QUERY;

} /* GW_SLP_Tool */




